import com.cycling74.max.*;
import java.io.*;
import java.net.*;

public class tcpClient extends MaxObject
{

	Socket dqdSocket = null;
	DataOutputStream os = null;
	BufferedReader is = null;

	String responseLine;
	String addressString = "localhost";
	int portSetting = 7126;
	boolean aThreadIsOpen = false;
	boolean serverConnected = false;
	int socketTimeout = 4000;

	private static final String[] INLET_ASSIST = new String[]{
		"messages in - ie open, close, address, port, timeout, send"
	};
	private static final String[] OUTLET_ASSIST = new String[]{
		"lines read back from TCP Server", "Connection Information"
	};

	private void startSocket() {
		try {
			dqdSocket = TimedSocket.getSocket(addressString,portSetting,socketTimeout);
			//dqdSocket = new Socket(addressString,portSetting);
			//dqdSocket.setSoTimeout(1000);
			os = new DataOutputStream(dqdSocket.getOutputStream());
			is = new BufferedReader(new InputStreamReader(dqdSocket.getInputStream()));
			serverConnected = true;
			post("Connected to " + addressString + ":" + portSetting);
			outlet(1,new Atom[]{Atom.newAtom("connected")});
		} catch (SocketTimeoutException e) {
			outlet(1,new Atom[]{Atom.newAtom("connection attempt timed out")});
		} catch (UnknownHostException e) {
			post("Don't know about host " + addressString + ":" + portSetting);
			outlet(1,new Atom[]{Atom.newAtom("dont know about host")});
		} catch (IOException e) {
			post("Couldn't get I/O for the connection to " + addressString + ":" + portSetting);
			outlet(1,new Atom[]{Atom.newAtom("cannot connect to host")});
		}
	}

	private void killSocket() {
		try {
			os.close();
			is.close();
			dqdSocket.close();
			serverConnected = false;
			outlet(1,new Atom[]{Atom.newAtom("connection closed")});
		}
		catch (IOException e) {
			post("IOException:  " + e);
		}
	}
	
	public tcpClient(Atom[] args)
	{
		declareInlets(new int[]{DataTypes.ALL});
		declareOutlets(new int[]{DataTypes.ALL,DataTypes.ALL});
		
		setInletAssist(INLET_ASSIST);
		setOutletAssist(OUTLET_ASSIST);
		declareAttribute("address", "getAddress", "setAddress");
		declareAttribute("port", "getPort", "setPort");
		declareAttribute("timeout", "getTimeout", "setTimeout");
		// need attribute for \n or \n\r or \r\n or whatever - default to just \n
	}


	private void setAddress(String s) {
		addressString = s;
	}

	private Atom[] getAddress() {
		return new Atom[] {Atom.newAtom("address " + addressString)};
	}
	
	private void setPort(int i) {
		portSetting = i;
	}
	
	private Atom[] getPort() {
		return new Atom[] {Atom.newAtom("port " + portSetting)};
	}
	
	private void setTimeout(int t) {
		socketTimeout = t;
	}
	
	private Atom[] getTimeout() {
		return new Atom[] {Atom.newAtom("timeout " + socketTimeout)};
	}
	
	public void open() {
		if (serverConnected) {
			close();
		}
		startSocket();
	}

	public void close() {
		if (dqdSocket != null && os != null && is != null) {
			killSocket();
		}
	}


	
	public void bang()
	{	
		if (!aThreadIsOpen && serverConnected) {
			//create a new thread to listen 
			Thread t = new Thread(){
				public void run()
				{
					aThreadIsOpen = true;
					if (dqdSocket != null && os != null && is != null) {
						try {
							responseLine = is.readLine();
						} catch (IOException e) {
							//System.err.println("Readback Thread IOException:  " + e);
							outlet(1,new Atom[]{Atom.newAtom("stopped listening on previous connection")});
							serverConnected = false;
							aThreadIsOpen = false;
							return;
						}
						if (responseLine == null) {
							outlet(1,new Atom[]{Atom.newAtom("server disconnected")});
							serverConnected = false;
						} else {
							//post(responseLine);
							outlet(0,new Atom[]{Atom.newAtom(responseLine)});
						}
					}
					aThreadIsOpen = false;
				}
			}; 
			t.start(); //start the thread executing
		}
	}
	
	public void inlet(int i)
	{
	}
	
	public void inlet(float f)
	{
	}
		
	//public void list(Atom[] list)
	//{
	
	//}

	public void send (String inputString)
	{
		if (dqdSocket != null && os != null && is != null) {
			try {
				os.writeBytes(inputString + "\n");
			} catch (UnknownHostException e) {
				post("Trying to connect to unknown host: " + e);
				outlet(1,new Atom[]{Atom.newAtom("unknown host")});
			} catch (IOException e) {
				//post("IOException:  " + e);
				outlet(1,new Atom[]{Atom.newAtom("no connection open")});	
			}
		} else {
			outlet(1,new Atom[]{Atom.newAtom("no connection open")});
		}
	}
		
}
